from typing import Dict

from bridger.metadata.mixins import BridgerViewSetConfig
from bridger.filters import DjangoFilterBackend
from bridger.utils import ilen

class FilterFieldsBridgerViewSetConfig(BridgerViewSetConfig):

    def get_metadata(self) -> str:
        for backend in getattr(self.view, "filter_backends", []):
            if isinstance(backend(), DjangoFilterBackend):
                filterset_class = backend().get_filterset_class(self.view, self.view.queryset)
                if filterset_class:
                    base_filters = getattr(filterset_class, "base_filters", {})
                    if filterset_class_meta := getattr(filterset_class, "Meta", None):
                        base_filters.update(getattr(filterset_class_meta, "df_fields", {}))
                    if base_filters:
                        field_names = [v.field_name for k, v in base_filters.items()]
                        stored_fields = dict()
                        for index, (name, field) in enumerate(base_filters.items()):

                            upcomming_fields = ilen(filter(lambda f: f == field.field_name, field_names[index + 1 :]))
                            representation = field.get_representation(self.request, name, self.view)

                            if upcomming_fields > 0 or field.field_name in stored_fields:
                                _representation = stored_fields.get(field.field_name, representation)
                                _representation["lookup_expr"].update(representation["lookup_expr"])
                                _representation["default"].update(representation["default"])
                                stored_fields[field.field_name] = _representation

                            if upcomming_fields == 0:
                                yield field.field_name, stored_fields.get(field.field_name, representation)

    @classmethod
    def get_metadata_key(cls):
        return "filter_fields"



"""
Provide Filters for Pandas based views
"""
import operator
from functools import reduce
from bridger import filters as wb_filters
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.template import loader
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from rest_framework.compat import coreapi, coreschema, distinct
from rest_framework.settings import api_settings
from rest_framework.filters import SearchFilter, OrderingFilter
from bridger.filters import DjangoFilterBackend

class PandasDjangoFilterBackend(DjangoFilterBackend):
    lookups_operator = {
        'lte': operator.le,
        'lt': operator.lt,
        'gte': operator.ge,
        'gt': operator.gt,
        'exact': operator.eq
    }
    def filter_dataframe(self, request, df, view):
        filterset_class = self.get_filterset_class(view, view.get_queryset())
        kwargs = self.get_filterset_kwargs(request, view.get_queryset(), view)

        filter_terms = filterset_class(**kwargs).form.data
        pandas_view_fields = view.pandas_fields.to_dict()
        conditions = []
        for filter_term, value in filter_terms.items():
            if _filter := getattr(filterset_class.Meta, "df_fields", {}).get(filter_term, None):
                # We support only number for now
                lookup_expr = getattr(_filter, "lookup_expr", 'exact')
                if isinstance(_filter, wb_filters.NumberFilter):
                    conditions.append(self.lookups_operator[lookup_expr](df[_filter.field_name], float(value)))
        if conditions:
            df = df[reduce(operator.and_, conditions)]
        return df

class PandasSearchFilter(SearchFilter):

    def filter_queryset(self, request, queryset, view):
        return queryset

    def filter_dataframe(self, request, df, view):
        search_fields = self.get_search_fields(view, request)
        search_terms = self.get_search_terms(request)
        if not search_fields or not search_terms or df.empty:
            return df

        search_fields = [field for field in search_fields if field in df.columns]
        conditions = []

        for search_term in search_terms:
            queries = [
                df[field].str.contains(search_term, na=False, case=False)
                for field in search_fields
            ]
            conditions.append(reduce(operator.or_, queries))
        df = df[reduce(operator.and_, conditions)]

        return df


class PandasOrderingFilter(OrderingFilter):

    def get_ordering_df(self, request, df, view):
        """
        Ordering is set by a comma delimited ?ordering=... query parameter.

        The `ordering` query parameter can be overridden by setting
        the `ordering_param` value on the OrderingFilter or by
        specifying an `ORDERING_PARAM` value in the API settings.
        """
        params = request.query_params.get(self.ordering_param)
        if params:
            fields = [param.strip() for param in params.split(',')]
            ordering = self.remove_invalid_fields_df(df, fields, view, request)
            if ordering:
                return ordering

        # No ordering was included, or all the ordering fields were invalid
        return self.get_default_ordering(view)

    def remove_invalid_fields_df(self, df, fields, view, request):
        valid_fields = getattr(view, 'ordering_fields', self.ordering_fields)

        def term_valid(term):
            if term.startswith("-"):
                term = term[1:]
            return term in df.columns and term in valid_fields

        return [term for term in fields if term_valid(term)]

    def filter_queryset(self, request, queryset, view):
        return queryset

    def filter_dataframe(self, request, df, view):
        base_ordering = self.get_ordering_df(request, df, view)
        if base_ordering:
            ordering_by = []
            ascending_list = []
            for order in base_ordering:
                ascending = order[0] != '-'
                ordering_by.append(order.replace('-', ''))
                ascending_list.append(ascending)

            if ordering_by and ascending_list:
                return df.sort_values(by=ordering_by, ascending=ascending_list)
        return df


import pandas as pd
from django.db.models import QuerySet
from rest_framework import filters
from rest_framework.filters import OrderingFilter
from rest_framework.response import Response
from rest_framework.views import APIView

from bridger.filters import DjangoFilterBackend
from bridger.metadata.views import MetadataMixin
from bridger.viewsets.mixins import DocumentationMixin, ModelMixin
from .metadata import PandasMetadata
import numpy as np
from .filters import PandasSearchFilter, PandasOrderingFilter, PandasDjangoFilterBackend

class PandasAPIView(MetadataMixin, DocumentationMixin, ModelMixin, APIView):

    filter_backends = [
        PandasOrderingFilter,
        PandasSearchFilter,
        PandasDjangoFilterBackend,
    ]

    metadata_class = PandasMetadata

    filterset_fields = {}
    search_fields = []
    ordering_fields = []

    def filter_queryset(self, queryset: QuerySet) -> QuerySet:
        for backend in list(self.filter_backends):
            queryset = backend().filter_queryset(self.request, queryset, self)
        return queryset

    def filter_dataframe(self, df):
        for backend in list(self.filter_backends):
            if hasattr(backend, "filter_dataframe"):
                df = backend().filter_dataframe(self.request, df, self)
        return df

    def get_queryset(self):
        assert hasattr(self, "queryset"), "Either specify a queryset or implement the get_queryset method."
        return self.queryset

    def get_dataframe(self, request, **kwargs):
        assert hasattr(self, "pandas_fields"), "No pandas_fields specified"
        queryset = self.get_queryset()
        if queryset.exists():
            queryset = self.filter_queryset(queryset)
            return pd.DataFrame(queryset.values(*self.pandas_fields.to_dict().keys()))
        return pd.DataFrame()

    def manipulate_dataframe(self, df):
        return df

    def get_aggregates(self, request, df):
        return {}

    def get(self, request, **kwargs):
        self.request = request
        df = self.manipulate_dataframe(self.get_dataframe(request, **kwargs))
        df = df.replace([np.inf, -np.inf], np.nan)
        df = df.where(pd.notnull(df), None)
        df = self.filter_dataframe(df)
        aggregates = self.get_aggregates(request, df) if not df.empty else {}
        return Response({"results": df.to_dict("records"), "aggregates": aggregates})


from bridger import buttons as bt
from bridger import display as dp
from bridger.buttons.enums import ButtonLevel
from bridger.utils.icons import WBIcon
from bridger.enums import RequestType
from bridger.buttons.metadata_config import ButtonConfig
from rest_framework.reverse import reverse

class NotificationButtonConfig(ButtonConfig):   
    
    def get_custom_buttons(self):
        if not self.view.kwargs.get("pk", None):
            return {
                bt.ActionButton(
                    method=RequestType.POST,
                    action_label="All notifications read.",
                    endpoint=reverse("bridger:notification-mark-all-as-read", request=self.request),
                    description_fields="Do you want to mark notifications as read?",
                    label="Mark all as read",
                    icon=WBIcon.EYE.value,
                    confirm_config=bt.ButtonConfig(label="Read all"),
                    cancel_config=bt.ButtonConfig(label="Cancel"),
                    identifiers=("relatedmodeltest-list",),
                ),
                bt.ActionButton(
                    method=RequestType.POST,
                    action_label="Delete all read notifications.",
                    endpoint=reverse("bridger:notification-delete-all-read", request=self.request),
                    description_fields="Do you want delete all read notifications?",
                    label="Delete all read notifications",
                    icon=WBIcon.TRASH.value,
                    confirm_config=bt.ButtonConfig(label="Delete all", level=bt.ButtonLevel.WARNING),
                    cancel_config=bt.ButtonConfig(label="Cancel", level=bt.ButtonLevel.ERROR),
                    identifiers=(reverse("bridger:notification-list", request=self.request),),
                ),
            }
        return dict()